import datetime
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import List, Tuple, Sequence, Literal

from rethink import const
from rethink.models.utils import strip_html_tags


@dataclass
class SearchDoc:
    nid: str
    title: str
    body: str

    def __post_init__(self):
        self.title = strip_html_tags(self.title)
        self.body = strip_html_tags(self.body)


@dataclass
class RestoreSearchDoc:
    nid: str
    title: str
    body: str
    createdAt: datetime.datetime
    modifiedAt: datetime.datetime
    disabled: bool
    inTrash: bool

    def __post_init__(self):
        self.title = strip_html_tags(self.title)
        self.body = strip_html_tags(self.body)


@dataclass
class SearchResult:
    nid: str
    score: float
    titleHighlight: str
    bodyHighlights: List[str]


STOPWORDS = sorted(list(set(
    (Path(__file__).parent / "cn_stopwords.txt").read_text(encoding="utf-8").splitlines()
    + (Path(__file__).parent / "baidu_stopwords.txt").read_text(encoding="utf-8").splitlines()
)))


class BaseEngine(ABC):
    hl_tag_name = "em"
    hl_class_name = "match"
    hl_term_prefix = "term"

    @abstractmethod
    async def init(self):
        ...

    @abstractmethod
    async def drop(self):
        ...

    @abstractmethod
    async def add(self, uid: str, doc: SearchDoc) -> const.Code:
        ...

    @abstractmethod
    async def update(self, uid: str, doc: SearchDoc) -> const.Code:
        ...

    @abstractmethod
    async def to_trash(self, uid: str, nid: str) -> const.Code:
        ...

    @abstractmethod
    async def restore_from_trash(self, uid: str, nid: str) -> const.Code:
        ...

    @abstractmethod
    async def disable(self, uid: str, nid: str) -> const.Code:
        ...

    @abstractmethod
    async def enable(self, uid: str, nid: str) -> const.Code:
        ...

    @abstractmethod
    async def delete(self, uid: str, nid: str) -> const.Code:
        ...

    @abstractmethod
    async def add_batch(self, uid: str, docs: List[SearchDoc]) -> const.Code:
        ...

    @abstractmethod
    async def delete_batch(self, uid: str, nids: List[str]) -> const.Code:
        ...

    @abstractmethod
    async def update_batch(self, uid: str, docs: List[SearchDoc]) -> const.Code:
        ...

    @abstractmethod
    async def batch_to_trash(self, uid: str, nids: List[str]) -> const.Code:
        ...

    @abstractmethod
    async def restore_batch_from_trash(self, uid: str, nids: List[str]) -> const.Code:
        ...

    @abstractmethod
    async def _search(
            self,
            uid: str,
            query: str = "",
            sort_key: str = None,
            reverse: bool = False,
            page: int = 1,
            page_size: int = 10,
            exclude_nids: Sequence[str] = None,
            with_stop_analyzer: bool = False,
    ):
        ...

    @abstractmethod
    async def search(
            self,
            uid: str,
            query: str = "",
            sort_key: Literal[
                "createdAt", "modifiedAt", "title", "similarity"
            ] = None,
            reverse: bool = False,
            page: int = 1,
            page_size: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> Tuple[List[SearchResult], int]:
        ...

    @abstractmethod
    async def recommend(
            self,
            uid: str,
            content: str,
            max_return: int = 10,
            exclude_nids: Sequence[str] = None,
    ) -> List[SearchResult]:
        ...

    @abstractmethod
    async def refresh(self):
        ...

    @abstractmethod
    async def count_all(self) -> int:
        ...

    @abstractmethod
    async def batch_restore_docs(self, uid: str, docs: List[RestoreSearchDoc]) -> const.Code:
        ...



import unittest

import elastic_transport

from rethink import const, config
from rethink.models.search_engine.engine_es import ESSearcher, SearchDoc
from . import utils


class ESTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.development")
        cls.searcher = ESSearcher()

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.development")

    async def asyncSetUp(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncSetUp skipped")
            return
        try:
            await self.searcher.drop()
            await self.searcher.init()
            self.assertTrue(await self.searcher.es.indices.exists(index=config.get_settings().ES_INDEX_ALIAS))
        except (elastic_transport.ConnectionError, RuntimeError):
            utils.skip_no_connect.skip = True
            print("remote test asyncSetUp timeout")
            if self.searcher.es is not None:
                await self.searcher.es.close()

    async def asyncTearDown(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncTearDown skipped")
            return
        try:
            await self.searcher.drop()
        except (elastic_transport.ConnectionError, RuntimeError):
            pass

    @utils.skip_no_connect
    async def test_reindex(self):
        code = await self.searcher.add_batch(uid="uid", docs=[SearchDoc(
            nid=f"nid{i}",
            title=f"title{i}",
            body=f"this is {i} doc, 这是第 {i} 个文档",
        ) for i in range(20)])
        self.assertEqual(const.Code.OK, code)
        await self.searcher.refresh()
        self.assertIn(config.get_settings().ES_INDEX_ALIAS, self.searcher.index)
        self.assertEqual("1", self.searcher.index.split("-")[-1])

        docs, total = await self.searcher.search(
            uid="uid",
            query="title doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)

        await self.searcher.reindex()
        await self.searcher.refresh()
        self.assertEqual(20, await self.searcher.count_all())
        self.assertIn(config.get_settings().ES_INDEX_ALIAS, self.searcher.index)
        self.assertEqual("2", self.searcher.index.split("-")[-1])

    @utils.skip_no_connect
    async def test_add(self):
        for i in range(20):
            code = await self.searcher.add(uid="uid", doc=SearchDoc(
                nid=f"nid{i}",
                title=f"title{i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ))
            self.assertEqual(const.Code.OK, code)

        await self.searcher.refresh()
        docs, total = await self.searcher.search(
            uid="uid",
            query="title doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )

        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)
        self.assertEqual("nid19", docs[0].nid)
        self.assertEqual("nid10", docs[-1].nid)
        self.assertEqual(['this is 19 <em class="match term0">doc</em>, 这是第 19 个文档'], docs[0].bodyHighlights)

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            page=200,
            page_size=10,
        )
        self.assertEqual(0, len(docs))
        self.assertEqual(20, total)

        docs, total = await self.searcher.search(
            uid="uid",
            query="",
            page=0,
            page_size=50,
        )
        self.assertEqual(20, len(docs))
        self.assertEqual(20, total)
        self.assertEqual([], docs[0].bodyHighlights)

        count = await self.searcher.count_all()
        self.assertEqual(20, count)

    @utils.skip_no_connect
    async def test_batch_add_update_delete(self):
        code = await self.searcher.add_batch(uid="uid", docs=[
            SearchDoc(
                nid=f"nid{i}",
                title=f"title{i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ) for i in range(20)
        ])
        self.assertEqual(const.Code.OK, code)
        await self.searcher.refresh()
        count = await self.searcher.count_all()
        self.assertEqual(20, count)

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)
        self.assertEqual("nid19", docs[0].nid)

        code = await self.searcher.update_batch(uid="uid", docs=[
            SearchDoc(
                nid=f"nid{i}",
                title=f"title_update{i}",
                body=f"this is {i} doc, 这是第 {i} 个文档",
            ) for i in range(20)
        ])
        self.assertEqual(const.Code.OK, code)
        await self.searcher.refresh()
        self.assertEqual(20, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(20, total)
        self.assertEqual("nid19", docs[0].nid)

        code = await self.searcher.disable(uid="uid", nid="nid18")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(20, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(19, total)
        self.assertEqual("nid17", docs[1].nid)

        code = await self.searcher.enable(uid="uid", nid="nid18")
        self.assertEqual(const.Code.OK, code)

        code = await self.searcher.batch_to_trash(uid="uid", nids=[f"nid{i}" for i in range(10)])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(20, await self.searcher.count_all())

        code = await self.searcher.delete_batch(uid="uid", nids=[f"nid{i}" for i in range(10)])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(10, await self.searcher.count_all())

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
        )
        self.assertEqual(10, len(docs))
        self.assertEqual(10, total)
        self.assertEqual("nid19", docs[0].nid)
        self.assertEqual("nid18", docs[1].nid)
        self.assertEqual("nid17", docs[2].nid)

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
            exclude_nids=[f"nid19"]
        )
        self.assertEqual(9, len(docs))
        self.assertEqual(9, total)
        self.assertEqual("nid18", docs[0].nid)

        docs, total = await self.searcher.search(
            uid="uid",
            query="doc",
            sort_key="title",
            reverse=False,
            page=0,
            page_size=10,
            exclude_nids=[f"nid19"]
        )
        self.assertEqual(9, len(docs))
        self.assertEqual(9, total)
        self.assertEqual("nid10", docs[0].nid)

    @utils.skip_no_connect
    async def test_multi_user(self):
        for uid in ["uid1", "uid2"]:
            code = await self.searcher.add_batch(uid=uid, docs=[
                SearchDoc(
                    nid=f"{uid}nid{i}",
                    title=f"title{i}",
                    body=f"this is {i} doc, 这是第 {i} 个文档",
                ) for i in range(20)
            ])
            self.assertEqual(const.Code.OK, code)

        await self.searcher.refresh()

        for uid in ["uid1", "uid2"]:
            docs, total = await self.searcher.search(
                uid=uid,
                query="doc",
                sort_key="createdAt",
                reverse=True,
                page=0,
                page_size=10,
            )
            self.assertEqual(10, len(docs))
            self.assertEqual(20, total)
            self.assertEqual(f"{uid}nid19", docs[0].nid)



from typing import Optional

from fastapi import APIRouter, Depends
from fastapi.responses import StreamingResponse

from rethink.controllers.verify import v_captcha
from rethink.routes.utils import measure_time_spend, verify_referer

router = APIRouter(
    prefix="/api",
    tags=["verification"],
    responses={404: {"description": "Not found"}},
)


@router.get(
    path="/captcha/img",
)
@measure_time_spend
async def get_captcha_img(
        referer: Optional[str] = Depends(verify_referer),
) -> StreamingResponse:
    return v_captcha.get_captcha_img()



import asyncio
import datetime
import multiprocessing
import os
import time
import zipfile
from typing import List, Optional, Tuple

import pymongo.errors
from bson import ObjectId
from bson.tz_util import utc

from rethink import const, models, config
from rethink.logger import logger
from rethink.models.files import file_ops
from rethink.models.tps import ImportData

RESIZE_IMG_THRESHOLD = 1024 * 256  # 256kb  # 1024 * 128  # 128 kb

ctx = multiprocessing.get_context('spawn')
QUEUE = ctx.Queue()


async def __set_running_false(
        uid: str,
        code: const.Code,
        msg: str = "",
) -> None:
    await models.database.COLL.import_data.update_one({"uid": uid}, {"$set": {
        "running": False,
        "msg": msg,
        "code": code.value,
    }})


async def update_process(
        uid: str,
        type_: str,
        process: int,
        start_at: datetime.datetime = None,
        running: bool = None,
        code: int = None,
) -> Tuple[Optional[ImportData], const.Code]:
    data = {"type": type_, "process": process}
    if start_at is not None:
        data["startAt"] = start_at
    if running is not None:
        data["running"] = running
    if code is not None:
        data["code"] = code
    if config.is_local_db():
        # local db not support find_one_and_update
        await models.database.COLL.import_data.update_one({"uid": uid}, {"$set": data})
        doc = await models.database.COLL.import_data.find_one({"uid": uid})
    else:
        doc = await models.database.COLL.import_data.find_one_and_update(
            {"uid": uid},
            {"$set": data}
        )
    if doc is None:
        return doc, const.Code.OPERATION_FAILED
    return doc, const.Code.OK


async def import_set_modules():
    from rethink import models

    await models.database.set_client()
    await models.database.searcher().init()
    models.database.set_coll()


async def __check_last_task_finished(uid: str, type_: str) -> Tuple[Optional[ImportData], bool]:
    doc = await models.database.COLL.import_data.find_one({"uid": uid})
    if doc and doc["running"]:
        await __set_running_false(
            uid=uid,
            code=const.Code.IMPORT_PROCESS_NOT_FINISHED,
            msg="last importData process not finished, please try again later",
        )
        return None, False

    if doc is None:
        doc: ImportData = {
            "_id": ObjectId(),
            "uid": uid,
            "process": 0,
            "type": type_,
            "startAt": datetime.datetime.now(tz=utc),
            "running": True,
            "msg": "",
            "code": 0,
            "obsidian": {},
        }
        res = await models.database.COLL.import_data.insert_one(doc)
        if not res.acknowledged:
            await __set_running_false(
                uid,
                const.Code.OPERATION_FAILED,
                msg="insert new importData process failed",
            )
            return doc, False
    else:
        doc, code = await update_process(
            uid=uid,
            type_=type_,
            process=0,
            start_at=datetime.datetime.now(tz=utc),
            running=True,
            code=0,
        )
        if code != const.Code.OK:
            await __set_running_false(
                uid=uid,
                code=code,
                msg="update importData process failed",
            )
            return doc, False
    return doc, True


async def __finish_task(uid: str, obsidian=None):
    resp = await models.database.COLL.import_data.update_one(
        {"uid": uid},
        {"$set": {
            "obsidian": obsidian or {},
            "running": False,
            "process": 100,
            "code": 0,
            "msg": "done",
        }})

    if resp.modified_count != 1:
        await __set_running_false(
            uid=uid,
            code=const.Code.OPERATION_FAILED,
            msg="uploading importData process failed",
        )


async def update_text_task(
        files: List[dict],
        max_file_size: int,
        uid: str,
):
    type_ = "md"
    await import_set_modules()

    doc, finished = await __check_last_task_finished(uid=uid, type_=type_)
    if not finished:
        return

    for file in files:
        if not file["filename"].endswith(".md") and not file["filename"].endswith(".txt"):
            await __set_running_false(
                uid=uid,
                code=const.Code.INVALID_FILE_TYPE,
                msg=f"invalid file type: {file['filename']}",
            )
            return
        if file["size"] > max_file_size:
            await __set_running_false(
                uid=uid,
                code=const.Code.TOO_LARGE_FILE,
                msg=f"file size: {file['size']} > {max_file_size} (max file size): {file['filename']}",
            )
            return

    for i, file in enumerate(files):
        try:
            md = file["content"].decode("utf-8")
        except (FileNotFoundError, OSError) as e:
            logger.error(f"error: {e}. filepath: {file['filename']}")
            await __set_running_false(
                uid=uid,
                code=const.Code.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {file['filename']}",
            )
            return
        title = file["filename"].rsplit(".", 1)[0]
        md = title + "\n\n" + md
        try:
            n, code = await models.node.add(
                uid=uid,
                md=md,
                type_=const.NodeType.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            continue
        except Exception as e:
            logger.error(f"error: {e}. filepath: {file['filename']}")
            await __set_running_false(
                uid=uid,
                code=const.Code.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {file['filename']}",
            )
            return
        if code != const.Code.OK:
            await __set_running_false(
                uid=uid,
                code=code,
                msg=f"file insert failed: {file['filename']}",
            )
            return
        if i % 20 == 0:
            doc, code = await update_process(uid=uid, type_=type_, process=int(i / len(files) * 100))
            if code != const.Code.OK:
                await __set_running_false(
                    uid=uid,
                    code=code,
                    msg="uploading process update failed",
                )
                return
            if not doc["running"]:
                break

    await __finish_task(uid=uid)


async def upload_obsidian_task(
        bytes_data: bytes,
        filename: str,
        max_file_size: int,
        uid: str,
) -> None:
    type_ = "obsidian"
    await import_set_modules()

    doc, finished = await __check_last_task_finished(uid=uid, type_=type_)
    if not finished:
        return

    t0 = time.time()
    try:
        unzipped_files = file_ops.unzip_file(bytes_data)
    except zipfile.BadZipFile as e:
        await __set_running_false(
            uid,
            const.Code.INVALID_FILE_TYPE,
            msg=f"unzip failed: {e}"
        )
        logger.info(f"invalid file type: {filename}, uid: {uid}")
        return
    t1 = time.time()
    logger.info(f"obsidian upload, uid={uid}, unzip time: {t1 - t0:.2f}")

    filtered_files = {}
    existed_filename2nid = doc.get("obsidian", {}).copy()
    img_path_dict = {}
    img_name_dict = {}
    md_count = 0
    for filepath, data in unzipped_files.items():
        try:
            base_name, ext = filepath.rsplit(".", 1)
        except ValueError:
            continue
        valid_file = {"md", "txt"}
        valid_file.update(file_ops.VALID_IMG_EXT)
        if ext not in valid_file:
            continue
        if data["size"] > max_file_size:
            await __set_running_false(
                uid,
                const.Code.TOO_LARGE_FILE,
                msg=f"file size > {max_file_size}: {filepath}",
            )
            logger.info(f"too large file: {filepath}, uid: {uid}")
            return
        if ext in ["md", "txt"]:
            if len(filepath.split("/")) > 1:
                continue
            md_count += 1
            filtered_files[filepath] = data["file"]
        else:
            img_path_dict[filepath] = data["file"]
            img_name_dict[os.path.basename(filepath)] = data["file"]
    t2 = time.time()
    logger.info(f"obsidian upload, uid={uid}, filter time: {t2 - t1:.2f}")

    # add new md files with only title
    for i, (filepath, file_bytes) in enumerate(filtered_files.items()):
        base_name, ext = filepath.rsplit(".", 1)
        if base_name in existed_filename2nid:
            continue
        try:
            n, code = await models.node.add(
                uid=uid,
                md=base_name,
                type_=const.NodeType.MARKDOWN.value,
            )
        except pymongo.errors.DuplicateKeyError:
            logger.error(f"duplicate key: {filepath}, uid: {uid}")
            continue
        if code != const.Code.OK:
            await __set_running_false(
                uid,
                code,
                msg=f"new file insert failed: {filepath}",
            )
            logger.error(f"error: {code}, filepath: {filepath}, uid: {uid}")
            return
        existed_filename2nid[base_name] = n["id"]
        if i % 20 == 0:
            doc, code = await update_process(uid=uid, type_=type_, process=int(i / md_count * 10))
            if code != const.Code.OK:
                await __set_running_false(
                    uid,
                    code,
                    msg="process updating failed",
                )
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
            if not doc["running"]:
                break
    t3 = time.time()
    logger.info(f"obsidian upload, uid={uid}, add new md time: {t3 - t2:.2f}")

    # update all files and update md files
    for i, (filepath, file_bytes) in enumerate(filtered_files.items()):
        base_name, ext = filepath.rsplit(".", 1)
        try:
            md = file_bytes.decode("utf-8")
        except (FileNotFoundError, OSError, UnicodeDecodeError) as e:
            logger.error(f"error: {e}. filepath: {filepath}")
            await __set_running_false(
                uid,
                const.Code.FILE_OPEN_ERROR,
                msg=f"file decode utf-8 failed, {e}: {filepath}",
            )
            logger.info(f"error: {const.Code.FILE_OPEN_ERROR}, filepath: {filepath}, uid: {uid}")
            return

        md = await file_ops.replace_inner_link_and_upload_image(
            uid=uid,
            md=md,
            exist_filename2nid=existed_filename2nid,
            img_path_dict=img_path_dict,
            img_name_dict=img_name_dict,
            resize_threshold=RESIZE_IMG_THRESHOLD,
        )
        md = base_name + "\n\n" + md
        nid = existed_filename2nid[base_name]
        n, code = await models.node.update(
            uid=uid,
            nid=nid,
            md=md,
            refresh_on_same_md=True,
        )
        if code == const.Code.NODE_NOT_EXIST:
            n, code = await models.node.add(
                uid=uid,
                md=md,
                type_=const.NodeType.MARKDOWN.value,
            )
            if code != const.Code.OK:
                await __set_running_false(
                    uid,
                    code,
                    msg=f"file insert failed: {filepath}",
                )
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
            existed_filename2nid[base_name] = n["id"]
        elif code != const.Code.OK:
            await __set_running_false(
                uid,
                code,
                msg=f"file updating failed: {filepath}",
            )
            logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
            return
        if i % 20 == 0:
            doc, code = await update_process(uid=uid, type_=type_, process=int(i / md_count * 40 + 10))
            if code != const.Code.OK:
                await __set_running_false(
                    uid,
                    code,
                    msg="uploading process update failed",
                )
                logger.info(f"error: {code}, filepath: {filepath}, uid: {uid}")
                return
            if not doc["running"]:
                break
    t4 = time.time()
    logger.info(f"obsidian upload, uid={uid}, update all files time: {t4 - t3:.2f}")

    # update for old obsidian files
    count = 0
    for base_name, nid in doc["obsidian"].items():
        if base_name not in existed_filename2nid:
            n, code = await models.node.get(uid=uid, nid=nid)
            if code != const.Code.OK:
                continue
            n, code = await models.node.update(
                uid=uid,
                nid=nid,
                md=n["md"],
                refresh_on_same_md=True,
            )
            if code != const.Code.OK:
                continue
        if count % 20 == 0:
            doc, code = await update_process(uid=uid, type_=type_, process=int(count / len(doc["obsidian"]) * 50 + 50))
            if code != const.Code.OK:
                await __set_running_false(
                    uid,
                    code,
                    msg="uploading process update failed",
                )
                logger.info(f"error: {code}, uid: {uid}")
                return
            if not doc["running"]:
                break

        count += 1

    t5 = time.time()
    logger.info(f"obsidian upload, uid={uid}, update for old obsidian files time: {t5 - t4:.2f}")

    await __finish_task(uid=uid, obsidian=existed_filename2nid)


def async_task(queue: multiprocessing.Queue):
    try:
        loop = asyncio.get_running_loop()
    except RuntimeError:
        loop = asyncio.get_event_loop()
    while True:
        item = queue.get()
        task = item.pop("task")
        if task == "md":
            func = update_text_task
        elif task == "obsidian":
            func = upload_obsidian_task
        else:
            logger.error(f"unknown task: {task}")
            continue
        try:
            loop.run_until_complete(func(**item))
        except Exception as e:
            logger.error(f"async task error: {e}")
    loop.close()


def init():
    p = ctx.Process(
        target=async_task,
        args=(QUEUE,),
        daemon=True,
    )
    p.start()



from typing import List, Optional

from pydantic import BaseModel, NonNegativeInt, Field

from rethink import const
from .search import NodesSearchResponse


class NodeData(BaseModel):
    class LinkedNode(BaseModel):
        id: str
        title: str
        md: str
        snippet: str
        type: NonNegativeInt
        disabled: bool
        inTrash: bool
        createdAt: str
        modifiedAt: str

    id: str
    md: str
    title: str
    snippet: str
    type: NonNegativeInt
    disabled: bool
    createdAt: str
    modifiedAt: str
    fromNodes: List[LinkedNode] = Field(default_factory=list)
    toNodes: List[LinkedNode] = Field(default_factory=list)


class PutRequest(BaseModel):
    md: str = Field(max_length=const.MD_MAX_LENGTH)
    type: NonNegativeInt
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)
    fromNid: str = Field(default="", max_length=const.NID_MAX_LENGTH)


class PutResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    node: Optional[NodeData]


class GetResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    node: Optional[NodeData]


class UpdateRequest(BaseModel):
    nid: str = Field(max_length=const.NID_MAX_LENGTH)
    md: str = Field(max_length=const.MD_MAX_LENGTH)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class RestoreFromTrashRequest(BaseModel):
    nid: str = Field(max_length=const.NID_MAX_LENGTH)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class GetFromTrashResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    data: NodesSearchResponse.Data


class BatchNodeIdsRequest(BaseModel):
    nids: List[str] = Field(default_factory=list, min_items=1, max_items=1000)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)



from random import randint
from typing import Tuple

from rethink import const
from rethink.controllers import schemas, auth
from rethink.logger import logger
from rethink.models.verify.email import email_server
from rethink.models.verify.verification import encode_numbers, verify_captcha


def __check_and_send_email(
        email: str,
        token: str,
        code_str: str,
        language: str,
) -> Tuple[str, int, const.Code]:
    code = verify_captcha(token=token, code_str=code_str)

    if code != const.Code.OK:
        return "", 0, code

    numbers = "".join([str(randint(0, 9)) for _ in range(6)])
    expired_min = 10
    code = email_server.send(
        recipient=email,
        numbers=numbers,
        expire=expired_min,
        language=language,
    )
    logger.info(f"send email to {email} with code {numbers}")
    return numbers, expired_min, code


def send_email_verification(req: schemas.user.EmailVerificationRequest) -> schemas.base.TokenResponse:
    if req.language not in const.Language.__members__:
        req.language = const.Language.EN.value
    numbers, expired_min, code = __check_and_send_email(
        email=req.email,
        token=req.captchaToken,
        code_str=req.captchaCode,
        language=req.language,
    )
    if code != const.Code.OK:
        return schemas.base.TokenResponse(
            code=code.value,
            message=const.get_msg_by_code(code, req.language),
            requestId=req.requestId,
        )

    token = encode_numbers(numbers=numbers, expired_min=expired_min)
    return schemas.base.TokenResponse(
        requestId=req.requestId,
        code=const.Code.OK.value,
        message=const.get_msg_by_code(const.Code.OK, req.language),
        token=token,
    )


async def check_email_then_send_email_verification(
        req: schemas.user.EmailVerificationRequest,
) -> schemas.base.TokenResponse:
    if req.language not in const.Language.__members__:
        req.language = const.Language.EN.value
    u, code = await auth.get_user_by_email(email=req.email)
    if code == const.Code.OK:
        code = const.Code.EMAIL_OCCUPIED
        return schemas.base.TokenResponse(
            code=code.value,
            message=const.get_msg_by_code(code, req.language),
            requestId=req.requestId,
        )
    return send_email_verification(req=req)



from rethink import const
from rethink.controllers import schemas, auth
from rethink.models.verify.verification import (
    verify_numbers,
)


async def reset_password(req: schemas.user.ResetPasswordRequest) -> schemas.base.AcknowledgeResponse:
    code = verify_numbers(token=req.verificationToken, number_str=req.verification)
    if code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=code.value,
            message=const.get_msg_by_code(code, const.Language.EN.value),
            requestId=req.requestId,
        )
    u, code = await auth.reset_password(
        email=req.email,
        password=req.newPassword,
    )

    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, u["language"]),
        requestId=req.requestId,
    )



import datetime
import io
from typing import List, Tuple, Optional

import httpx
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from rethink import const, models
from rethink.config import is_local_db
from rethink.logger import logger
from . import file_ops, queuing_upload

MAX_IMAGE_SIZE = 1024 * 1024 * 10  # 10 mb
QUEUE_INITED = False


async def upload_obsidian(uid: str, zipped_files: List[UploadFile]) -> const.Code:
    max_file_count = 1
    max_file_size = 1024 * 1024 * 200  # 200 mb

    if len(zipped_files) > max_file_count:
        return const.Code.TOO_MANY_FILES

    zipped_file = zipped_files[0]
    if not zipped_file.filename.endswith(".zip"):
        return const.Code.INVALID_FILE_TYPE

    if zipped_file.content_type not in ["application/zip", "application/octet-stream", "application/x-zip-compressed"]:
        return const.Code.INVALID_FILE_TYPE

    bytes_data = zipped_file.file.read()
    filename = zipped_file.filename

    if is_local_db():
        # local db not support find_one_and_update
        await queuing_upload.upload_obsidian_task(
            bytes_data=bytes_data,
            filename=filename,
            max_file_size=max_file_size,
            uid=uid,
        )
    else:
        global QUEUE_INITED
        if not QUEUE_INITED:
            queuing_upload.init()
            QUEUE_INITED = True
        queuing_upload.QUEUE.put_nowait({
            "task": "obsidian",
            "bytes_data": bytes_data,
            "filename": filename,
            "max_file_size": max_file_size,
            "uid": uid,
        })
    return const.Code.OK


async def upload_text(uid: str, files: List[UploadFile]) -> const.Code:
    max_file_count = 200
    max_file_size = 1024 * 512  # 512 kb

    doc = await models.database.COLL.import_data.find_one({"uid": uid})
    if doc is not None and doc["running"]:
        return const.Code.IMPORT_PROCESS_NOT_FINISHED

    if len(files) > max_file_count:
        return const.Code.TOO_MANY_FILES

    file_list = [{
        "filename": file.filename,
        "content": file.file.read(),
        "size": file.size,
    } for file in files]

    if is_local_db():
        # local db not support find_one_and_update
        await queuing_upload.update_text_task(
            files=file_list,
            max_file_size=max_file_size,
            uid=uid,
        )
    else:
        global QUEUE_INITED
        if not QUEUE_INITED:
            queuing_upload.init()
            QUEUE_INITED = True
        queuing_upload.QUEUE.put_nowait({
            "task": "md",
            "files": file_list,
            "max_file_size": max_file_size,
            "uid": uid,
        })
    return const.Code.OK


async def get_upload_process(uid: str) -> Optional[dict]:
    timeout_minus = 5
    doc = await models.database.COLL.import_data.find_one({"uid": uid})
    if doc is None:
        return None
    now = datetime.datetime.now(tz=utc)

    # upload timeout
    if doc["running"] and \
            now.replace(tzinfo=None) - doc["startAt"].replace(tzinfo=None) \
            > datetime.timedelta(minutes=timeout_minus):
        doc["running"] = False
        await models.database.COLL.import_data.update_one(
            {"uid": uid},
            {"$set": {
                "running": False,
                "code": const.Code.UPLOAD_TASK_TIMEOUT.value,
                "msg": f"Timeout, upload not finish in {timeout_minus} mins",
            }},
        )
    return doc


async def upload_image_vditor(uid: str, files: List[UploadFile]) -> dict:
    res = {
        "errFiles": [],
        "succMap": {},
        "code": const.Code.OK,
    }
    u, code = await models.user.get(uid=uid)
    if code != const.Code.OK:
        res["errFiles"] = [file.filename for file in files]
        res["code"] = code
        return res
    if await models.user.user_space_not_enough(u=u):
        res["errFiles"] = [file.filename for file in files]
        res["code"] = const.Code.USER_SPACE_NOT_ENOUGH
        return res

    return await file_ops.save_upload_files(
        uid=uid,
        files=files,
        max_image_size=MAX_IMAGE_SIZE,
        resize_threshold=queuing_upload.RESIZE_IMG_THRESHOLD,
    )


async def fetch_image_vditor(uid: str, url: str) -> Tuple[str, const.Code]:
    u, code = await models.user.get(uid=uid)
    if code != const.Code.OK:
        return "", code
    if await models.user.user_space_not_enough(u=u):
        return "", const.Code.USER_SPACE_NOT_ENOUGH
    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                url=url,
                headers=models.utils.ASYNC_CLIENT_HEADERS,
                timeout=5.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.info(f"failed to get {url}: {e}")
            return "", const.Code.FILE_OPEN_ERROR
        if response.status_code != 200:
            return "", const.Code.FILE_OPEN_ERROR

        content = response.content

        file = UploadFile(
            filename=url.split("/")[-1],
            file=io.BytesIO(content),
            headers=Headers(response.headers),
            size=len(content)
        )

    res = await file_ops.save_upload_files(
        uid=uid,
        files=[file],
        max_image_size=MAX_IMAGE_SIZE,
        resize_threshold=queuing_upload.RESIZE_IMG_THRESHOLD,
    )
    if len(res["errFiles"]) > 0:
        return "", const.Code.FILE_OPEN_ERROR
    return res["succMap"][file.filename], const.Code.OK



import datetime
import shutil
import unittest
from io import BytesIO
from pathlib import Path
from textwrap import dedent
from unittest.mock import patch

import httpx
from PIL import Image
from bson import ObjectId
from bson.tz_util import utc
from fastapi import UploadFile
from starlette.datastructures import Headers

from rethink import const, models
from rethink.models.files import queuing_upload
from rethink.models.utils import short_uuid
from . import utils


class LocalModelsTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")
        # create a fake image
        tmp_path = Path(__file__).parent / "tmp" / "fake.png"
        tmp_path.parent.mkdir(parents=True, exist_ok=True)
        Image.new("RGB", (100, 100)).save(tmp_path)

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")
        shutil.rmtree(Path(__file__).parent / "tmp", ignore_errors=True)

    async def asyncSetUp(self) -> None:
        await models.database.drop_all()
        await models.database.init()
        u, _ = await models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.uid = u["id"]

    async def asyncTearDown(self) -> None:
        await models.database.drop_all()
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "files", ignore_errors=True)
        shutil.rmtree(Path(__file__).parent / "tmp" / ".data" / "md", ignore_errors=True)

    async def test_user(self):
        _id, code = await models.user.add(
            account="aaa", source=const.UserSource.EMAIL.value,
            email="aaa", hashed="bbb", nickname="ccc", avatar="ddd", language=const.Language.EN.value)
        self.assertNotEqual("", _id)
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = await models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("ccc", u["nickname"])

        u, code = await models.user.update(
            uid=_id,
            hashed="1",
            nickname="2",
            avatar="3",
            language="zh",
            node_display_method=const.NodeDisplayMethod.LIST.value,
        )
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("1", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])
        self.assertEqual("zh", u["language"])
        self.assertEqual(const.NodeDisplayMethod.LIST.value, u["lastState"]["nodeDisplayMethod"])

        code = await models.user.disable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get(uid=_id)
        self.assertEqual(const.Code.ACCOUNT_OR_PASSWORD_ERROR, code)
        self.assertIsNone(u)

        code = await models.user.enable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        code = await models.user.disable(uid="sdwqdqw")
        self.assertEqual(const.Code.OPERATION_FAILED, code)

        code = await models.user.delete(uid=_id)
        self.assertEqual(const.Code.OK, code)

    async def test_node(self):
        node, code = await models.node.add(
            uid=self.uid, md="a" * (const.MD_MAX_LENGTH + 1), type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.NOTE_EXCEED_MAX_LENGTH, code)
        self.assertIsNone(node)

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        used_space = u["usedSpace"]
        node, code = await models.node.add(
            uid=self.uid, md="[title](/qqq)\nbody", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + len(node["md"].encode("utf-8")), u["usedSpace"])
        self.assertEqual("modifiedAt", u["lastState"]["nodeDisplaySortKey"])

        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title", n["title"])
        self.assertEqual("body", n["snippet"])

        ns, total = await models.database.searcher().search(uid=self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        ns, total = await models.database.searcher().search(uid=self.uid, page_size=5, page=12, sort_key="createdAt")
        self.assertEqual(0, len(ns))
        self.assertEqual(3, total)

        u, code = await models.user.update(
            uid=self.uid,
            node_display_sort_key="createdAt",
        )

        self.assertEqual(const.Code.OK, code)
        self.assertEqual("createdAt", u["lastState"]["nodeDisplaySortKey"])
        used_space = u["usedSpace"]
        n, code = await models.node.update(uid=self.uid, nid=node["id"], md="title2\nbody2")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("title2\nbody2", n["md"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + (
                len(n["md"].encode("utf-8")) -
                len(node["md"].encode("utf-8"))
        ), u["usedSpace"])

        code = await models.node.disable(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        code = await models.node.to_trash(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)

        code = await models.node.delete(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space - len(node["md"].encode("utf-8")), u["usedSpace"])

    async def test_parse_at(self):
        nid1, _ = await models.node.add(
            uid=self.uid, md="c", type_=const.NodeType.MARKDOWN.value,
        )
        nid2, _ = await models.node.add(
            uid=self.uid, md="我133", type_=const.NodeType.MARKDOWN.value,
        )
        md = dedent(f"""title
        fffqw [@c](/n/{nid1['id']})
        fff
        [@我133](/n/{nid2['id']})
        ffq
        """)
        node, code = await models.node.add(
            uid=self.uid, md=md, type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        nodes, total = await models.database.searcher().search(
            uid=self.uid,
            query="",
            sort_key="createdAt",
            reverse=True,
            page=0,
            page_size=10,
            exclude_nids=[],
        )
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)
        found, total = await models.database.searcher().search(uid=self.uid, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(2, total)

        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        cache = n["md"]
        n, code = await models.node.update(uid=self.uid, nid=node["id"], md=f'{cache}xxxx')
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(cache + "xxxx", n["md"])

        n, code = await models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, code = await models.node.update(uid=self.uid, nid=node["id"], md=n["title"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = await models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    async def test_add_set(self):
        node, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.Code.OK, code)

        res = await models.db_ops.node_add_to_set(node["id"], "toNodeIds", short_uuid())
        self.assertEqual(1, res.modified_count)
        node, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    async def test_cursor_text(self):
        n1, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = await models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        recom, total = await models.search.cursor_query(
            uid=self.uid,
            nid=n2["id"],
            query="text",
            page=0,
            page_size=10,
        )
        self.assertEqual(1, len(recom))
        self.assertEqual(1, total)

        recom, total = await models.search.cursor_query(
            uid=self.uid,
            nid=n2["id"],  # exclude the second node
            query="",  # return recent nodes only
            page=0,
            page_size=10,
        )
        self.assertEqual(2, len(recom))
        self.assertEqual(2, total)

        code = await models.search.add_recent_cursor_search(self.uid, n1["id"], n2["id"])
        self.assertEqual(const.Code.OK, code)

        recom, total = await models.search.cursor_query(
            uid=self.uid,
            nid=n1["id"],  # exclude the second node
            query="",
            page=0,
            page_size=10,
        )
        self.assertEqual(3, len(recom))
        self.assertEqual(3, total)
        self.assertEqual("Welcome to Rethink", recom[2].title)

    async def test_to_trash(self):
        n1, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = await models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        code = await models.node.to_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        ns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        ns, total = await models.database.searcher().search(self.uid, query="")
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = await models.node.restore_from_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)
        nodes, total = await models.database.searcher().search(self.uid, query="")
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    async def test_search(self):
        code = await models.search.put_recent_search(self.uid, "a")
        self.assertEqual(const.Code.OK, code)
        await models.search.put_recent_search(self.uid, "c")
        await models.search.put_recent_search(self.uid, "b")

        queries = await models.search.get_recent_search(self.uid)
        self.assertEqual(["b", "c", "a"], queries)

    async def test_batch(self):
        ns = []
        for i in range(10):
            n, code = await models.node.add(
                uid=self.uid, md=f"title{i}\ntext", type_=const.NodeType.MARKDOWN.value
            )
            self.assertEqual(const.Code.OK, code)
            ns.append(n)

        base_count = 2

        code = await models.node.batch_to_trash(self.uid, [n["id"] for n in ns[:4]])
        self.assertEqual(const.Code.OK, code)
        nodes, total = await models.database.searcher().search(self.uid, query="")
        self.assertEqual(6 + base_count, len(nodes))
        self.assertEqual(6 + base_count, total)

        tns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(4, total)
        self.assertEqual(4, len(tns))

        code = await models.node.restore_batch_from_trash(self.uid, [n["id"] for n in tns[:2]])
        self.assertEqual(const.Code.OK, code)
        nodes, total = await models.database.searcher().search(self.uid)
        self.assertEqual(8 + base_count, len(nodes))
        self.assertEqual(8 + base_count, total)

        code = await models.node.batch_delete(self.uid, [n["id"] for n in tns[2:4]])
        self.assertEqual(const.Code.OK, code)
        tns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(0, total)
        self.assertEqual(0, len(tns))

    async def test_files_upload_process(self):
        now = datetime.datetime.now(tz=utc)
        doc: models.tps.ImportData = {
            "_id": ObjectId(),
            "uid": "xxx",
            "process": 0,
            "type": "text",
            "startAt": now,
            "running": True,
            "obsidian": {},
            "msg": "",
            "code": const.Code.OK.value,
        }
        res = await models.database.COLL.import_data.insert_one(doc)
        self.assertTrue(res.acknowledged)

        doc, code = await queuing_upload.update_process("xxx", "obsidian", 10)
        self.assertEqual(const.Code.OK, code)

        doc = await models.files.get_upload_process("xxx")
        self.assertEqual(10, doc["process"])
        self.assertEqual("obsidian", doc["type"])
        self.assertEqual(now, doc["startAt"])
        self.assertTrue(doc["running"])

        await models.database.COLL.import_data.delete_one({"uid": "xxx"})

    async def test_update_title_and_from_nodes_updates(self):
        n1, code = await models.node.add(
            uid=self.uid, md="title1\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = await models.node.add(
            uid=self.uid, md=f"title2\n[@title1](/n/{n1['id']})", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        n1, code = await models.node.update(uid=self.uid, nid=n1["id"], md="title1Changed\ntext")
        self.assertEqual(const.Code.OK, code)
        n2, code = await models.node.get(uid=self.uid, nid=n2["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(f"title2\n[@title1Changed](/n/{n1['id']})", n2["md"])

    async def test_upload_image_vditor(self):
        u, code = await models.user.get(self.uid)
        used_space = u["usedSpace"]
        p = Path(__file__).parent / "tmp" / "fake.png"

        image = Image.open(p)
        buf = BytesIO()
        image.save(buf, format="png")
        size = buf.tell()
        img_file = UploadFile(
            buf, filename="fake.png", size=size,
            headers=Headers({"content-type": "image/png"})
        )
        res = await models.files.upload_image_vditor(self.uid, [img_file])
        self.assertIn("fake.png", res["succMap"])
        self.assertTrue(".png" in res["succMap"]["fake.png"])
        local_file = Path(__file__).parent / "tmp" / ".data" / res["succMap"]["fake.png"][1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()
        image.close()
        buf.close()
        await img_file.close()

        u, code = await models.user.get(self.uid)
        self.assertEqual(used_space + size, u["usedSpace"])

    @patch(
        "rethink.models.files.upload.httpx.AsyncClient.get",
    )
    async def test_fetch_image_vditor(self, mock_get):
        f = open(Path(__file__).parent / "tmp" / "fake.png", "rb")
        mock_get.return_value = httpx.Response(
            200,
            content=f.read(),
            headers={"content-type": "image/png"}
        )

        u, code = await models.user.get(self.uid)
        used_space = u["usedSpace"]

        url = "https://rethink.run/favicon.png"
        new_url, code = await models.files.fetch_image_vditor(self.uid, url)
        self.assertEqual(const.Code.OK, code)
        self.assertTrue(new_url.endswith(".png"))
        self.assertTrue(new_url.startswith("/"))
        local_file = Path(__file__).parent / "tmp" / ".data" / new_url[1:]
        self.assertTrue(local_file.exists())
        local_file.unlink()

        u, code = await models.user.get(self.uid)
        self.assertEqual(used_space + f.tell(), u["usedSpace"])
        f.close()

    async def test_update_used_space(self):
        u, code = await models.user.get(self.uid)
        base_used_space = u["usedSpace"]
        for delta, value in [
            (100, 100),
            (100, 200),
            (0, 200),
            (-3000, 0),
            (20.1, 20.1),
        ]:
            code = await models.user.update_used_space(self.uid, delta)
            self.assertEqual(const.Code.OK, code)
            u, code = await models.user.get(self.uid)
            self.assertEqual(const.Code.OK, code)
            now = u["usedSpace"] - base_used_space
            if now < 0:
                now = 0
                base_used_space = 0
            self.assertAlmostEqual(value, now, msg=f"delta: {delta}, value: {value}")



from typing import Optional

from fastapi import APIRouter, Depends

from rethink.controllers import schemas, email
from rethink.routes.utils import measure_time_spend, verify_referer

router = APIRouter(
    prefix="/api",
    tags=["email"],
    responses={404: {"description": "Not found"}},
)


@router.post(
    path="/email/verify",
    response_model=schemas.base.TokenResponse,
)
@measure_time_spend
async def email_verification(
        req: schemas.user.EmailVerificationRequest,
        referer: Optional[str] = Depends(verify_referer),
) -> schemas.base.TokenResponse:
    return email.send_email_verification(req=req)


@router.post(
    path="/email/register",
    response_model=schemas.base.TokenResponse,
)
@measure_time_spend
async def email_register(
        req: schemas.user.EmailVerificationRequest,
        referer: Optional[str] = Depends(verify_referer),
) -> schemas.base.TokenResponse:
    return await email.check_email_then_send_email_verification(req=req)



import re

# at least 6 characters, at most 20 characters, at least one letter and one number
VALID_PASSWORD_PTN = re.compile(r"^(?=.*[A-Za-z])(?=.*\d).{6,20}$")

EMAIL_PTN = re.compile(r'^[.0-9a-zA-Z_]{1,18}@([0-9a-zA-Z-]{1,13}\.){1,}[a-zA-Z]{1,3}$')



import os

from fastapi import FastAPI, HTTPException
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import HTMLResponse, FileResponse
from fastapi.staticfiles import StaticFiles
from starlette.middleware.base import BaseHTTPMiddleware

from rethink import const, config
from rethink.logger import logger, add_rotating_file_handler
from .models import database
from .routes import (
    user,
    oauth,
    node,
    search,
    trash,
    verification,
    files,
    email,
)

app = FastAPI(
    docs_url="/docs",
    openapi_url="/openapi.json",
)

if os.getenv("VUE_APP_MODE", "local") not in ["development", "local"]:
    allow_origins = [
        "https://rethink.run",
        "https://www.rethink.run",
    ]
    csp_local = ""
else:
    allow_origins = [
        "*",
    ]
    csp_local = " http://localhost:* http://127.0.0.1:* "


class CSPMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        csp_header = "default-src 'self'; " \
                     f"script-src 'self' 'unsafe-inline' 'unsafe-eval' {csp_local} https://www.googletagmanager.com; " \
                     f"style-src 'self' 'unsafe-inline' 'report-sample' {csp_local}; " \
                     "img-src * data: blob:; " \
                     f"connect-src 'self' {csp_local} https://www.google-analytics.com;"
        response.headers["Content-Security-Policy"] = csp_header
        return response


class FrameOptionsMiddleware(BaseHTTPMiddleware):
    async def dispatch(self, request, call_next):
        response = await call_next(request)
        response.headers["X-Frame-Options"] = "SAMEORIGIN"
        return response


app.add_middleware(
    CORSMiddleware,
    allow_origins=allow_origins,
    allow_credentials=True,
    allow_methods=["*"],
    allow_headers=["*"],
    expose_headers=["X-Captcha-Token"],
)
app.add_middleware(CSPMiddleware)
app.add_middleware(FrameOptionsMiddleware)

app.include_router(node.router)
app.include_router(user.router)
app.include_router(oauth.router)
app.include_router(search.router)
app.include_router(trash.router)
app.include_router(verification.router)
app.include_router(files.router)
app.include_router(email.router)


@app.on_event("startup")
async def startup_event():
    if not config.is_local_db():
        add_rotating_file_handler(
            log_dir=const.RETHINK_DIR.parent.parent / "logs",
            max_bytes=10 * 1024 * 1024,
            backup_count=10,
        )
    logger.debug(f'startup_event LOCAL_STORAGE_PATH: {os.environ.get("LOCAL_STORAGE_PATH")}')
    logger.debug(f'startup_event VUE_APP_MODE: {os.environ.get("VUE_APP_MODE")}')
    logger.debug(f'startup_event VUE_APP_API_PORT: {os.environ.get("VUE_APP_API_PORT")}')
    logger.debug(f'startup_event VUE_APP_LANGUAGE: {os.environ.get("VUE_APP_LANGUAGE")}')
    await database.init()
    logger.info("db initialized")


@app.on_event("shutdown")
async def shutdown_event():
    try:
        await database.searcher().es.close()
    except (AttributeError, ValueError):
        pass
    logger.info("db closed")


try:
    for name in ["css", "js", "img", "dist"]:
        app.mount(
            f"/{name}",
            StaticFiles(directory=const.FRONTEND_DIR / name),
            name=name,
        )
except RuntimeError:
    logger.info("mount frontend failed")


@app.get("/", response_class=HTMLResponse)
@app.get("/login", response_class=HTMLResponse)
@app.get("/about", response_class=HTMLResponse)
@app.get("/r", response_class=HTMLResponse)
@app.get("/r/{path}", response_class=HTMLResponse)
@app.get("/n/{nid}", response_class=HTMLResponse)
async def index() -> HTMLResponse:
    content = (const.FRONTEND_DIR / "index.html").read_text(encoding="utf-8")
    content += f"<script>window.VUE_APP_API_PORT = {os.getenv('VUE_APP_API_PORT')}</script>"
    return HTMLResponse(
        content=content,
        status_code=200,
    )


@app.get("/files/{fid}", response_class=FileResponse)
async def user_data(
        fid: str,
) -> FileResponse:
    if config.is_local_db():
        prefix = config.get_settings().LOCAL_STORAGE_PATH
    else:
        raise HTTPException(status_code=404, detail="only support local storage")
    return FileResponse(
        path=prefix / ".data" / "files" / fid,
        status_code=200,
    )



import calendar
import datetime
import json
import os
import struct
from dataclasses import dataclass
from typing import Optional, Union, TYPE_CHECKING

from bson import ObjectId
from bson.tz_util import utc

from rethink import config, const
from rethink.logger import logger
from rethink.mongita import MongitaClientDisk
from rethink.mongita.collection import Collection
from . import utils
from .search_engine.engine import BaseEngine, SearchDoc, RestoreSearchDoc
from .search_engine.engine_local import LocalSearcher
from .tps import UserMeta, Node, UserFile, ImportData

if TYPE_CHECKING:
    from motor.motor_asyncio import AsyncIOMotorClient, AsyncIOMotorCollection
    from .search_engine.engine_es import ESSearcher


@dataclass
class Collections:
    users: Union[Collection, "AsyncIOMotorCollection"] = None
    nodes: Union[Collection, "AsyncIOMotorCollection"] = None
    import_data: Union[Collection, "AsyncIOMotorCollection"] = None
    user_file: Union[Collection, "AsyncIOMotorCollection"] = None


COLL = Collections()
CLIENT: Optional[Union["AsyncIOMotorClient", MongitaClientDisk]] = None
SEARCHER: Optional[BaseEngine] = None


def searcher() -> BaseEngine:
    if SEARCHER is None:
        raise ValueError("searcher not initialized")
    return SEARCHER


async def set_client():
    global CLIENT, SEARCHER
    conf = config.get_settings()
    if config.is_local_db():
        if not conf.LOCAL_STORAGE_PATH.exists():
            raise FileNotFoundError(f"Path not exists: {conf.LOCAL_STORAGE_PATH}")
        db_path = conf.LOCAL_STORAGE_PATH / ".data" / "db"
        db_path.mkdir(parents=True, exist_ok=True)
        if not isinstance(SEARCHER, LocalSearcher):
            SEARCHER = LocalSearcher()
        CLIENT = MongitaClientDisk(db_path)
    else:
        from .search_engine.engine_es import ESSearcher
        if not isinstance(SEARCHER, ESSearcher):
            SEARCHER = ESSearcher()
        from motor.motor_asyncio import AsyncIOMotorClient
        CLIENT = AsyncIOMotorClient(
            host=conf.DB_HOST,
            port=conf.DB_PORT,
            username=conf.DB_USER,
            password=conf.DB_PASSWORD,
            socketTimeoutMS=1000 * 5,
        )


def set_coll():
    db = CLIENT[config.get_settings().DB_NAME]
    COLL.users = db["users"]
    COLL.nodes = db["nodes"]
    COLL.import_data = db["importData"]
    COLL.user_file = db["userFile"]


async def init():
    await set_client()
    await SEARCHER.init()
    set_coll()

    if config.is_local_db():
        await __local_try_create_or_restore()

    else:
        # CLIENT.get_io_loop = asyncio.get_running_loop
        await __remote_try_build_index()

    await __try_restore_search()


async def drop_all():
    await set_client()
    await CLIENT.drop_database(config.get_settings().DB_NAME)
    global SEARCHER
    if SEARCHER:
        await SEARCHER.drop()


async def __remote_try_build_index():
    # try creating index
    from motor.motor_asyncio import AsyncIOMotorClient
    if not isinstance(CLIENT, AsyncIOMotorClient):
        return

    users_info = await COLL.users.index_information()
    if "id_1" not in users_info:
        await COLL.users.create_index("id", unique=True)
    if "account_1_source_1" not in users_info:
        await COLL.users.create_index(["account", "source"], unique=True)

    nodes_info = await COLL.nodes.index_information()
    if "id_1" not in nodes_info:
        await COLL.nodes.create_index("id", unique=True)
    if "uid_1_id_-1" not in nodes_info:
        # created at
        await COLL.nodes.create_index(
            [("uid", 1), ("id", -1)],
            unique=True,
        )

    import_data_info = await COLL.import_data.index_information()
    if "uid_1" not in import_data_info:
        await COLL.import_data.create_index("uid", unique=True)

    user_file_info = await COLL.user_file.index_information()
    if "uid_1_fid_-1" not in user_file_info:
        await COLL.user_file.create_index([("uid", 1), ("fid", -1)], unique=True)


async def __local_try_add_default_user():
    dot_rethink_path = config.get_settings().LOCAL_STORAGE_PATH / ".data" / ".rethink.json"
    u_insertion = {
        "_id": ObjectId(),
        "id": utils.short_uuid(),
        "email": const.DEFAULT_USER["email"],
        "nickname": const.DEFAULT_USER["nickname"],
        "avatar": const.DEFAULT_USER["avatar"],
        "account": const.DEFAULT_USER["email"],
        "language": os.getenv("VUE_APP_LANGUAGE", const.Language.EN.value),
    }
    with open(dot_rethink_path, "w", encoding="utf-8") as f:
        out = u_insertion.copy()
        out["_id"] = str(out["_id"])
        json.dump(out, f, indent=2, ensure_ascii=False)

    logger.info("running at the first time, a user with initial data will be created")
    ns = const.NEW_USER_DEFAULT_NODES[u_insertion["language"]]
    search_docs = []

    async def create_node(md: str, to_nid: Optional[str] = None):
        title_, body_, snippet_ = utils.preprocess_md(md)
        n: Node = {
            "_id": ObjectId(),
            "id": utils.short_uuid(),
            "uid": u_insertion["id"],
            "title": title_,
            "snippet": snippet_,
            "md": md,
            "type": const.NodeType.MARKDOWN.value,
            "disabled": False,
            "inTrash": False,
            "modifiedAt": datetime.datetime.now(tz=utc),
            "inTrashAt": None,
            "fromNodeIds": [],
            "toNodeIds": [] if to_nid is None else [to_nid],
        }
        res = await COLL.nodes.insert_one(n)
        if not res.acknowledged:
            raise ValueError("cannot insert default node")
        search_docs.append(
            SearchDoc(
                nid=n["id"],
                title=title_,
                body=body_,
            )
        )
        md_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "md"
        md_dir.mkdir(parents=True, exist_ok=True)
        with open(md_dir / f"{n['id']}.md", "w", encoding="utf-8") as f:
            f.write(md)
        return n

    n0 = await create_node(ns[0])
    _md = ns[1].format(n0["id"])
    n1 = await create_node(_md, to_nid=n0["id"])

    u: UserMeta = {
        "_id": ObjectId(u_insertion["_id"]),
        "id": u_insertion["id"],
        "email": u_insertion["email"],
        "nickname": u_insertion["nickname"],
        "avatar": u_insertion["avatar"],
        "account": u_insertion["email"],
        "language": u_insertion["language"],

        "source": const.UserSource.LOCAL.value,
        "hashed": "",
        "disabled": False,
        "modifiedAt": datetime.datetime.now(tz=utc),
        "usedSpace": 0,
        "type": const.USER_TYPE.NORMAL.id,
        "lastState": {
            "recentCursorSearchSelectedNIds": [n0["id"], n1["id"]],
            "recentSearch": [],
            "nodeDisplayMethod": const.NodeDisplayMethod.CARD.value,
            "nodeDisplaySortKey": "modifiedAt"
        }
    }
    _ = await COLL.users.insert_one(u)

    await SEARCHER.add_batch(
        uid=u["id"],
        docs=search_docs,
    )


def __oid_from_datetime(dt: datetime.datetime) -> ObjectId:
    offset = dt.utcoffset()
    if offset is not None:
        dt = dt - offset
    timestamp = calendar.timegm(dt.timetuple())
    oid = struct.pack(">I", timestamp)
    oid += ObjectId._random()
    with ObjectId._inc_lock:
        oid += struct.pack(">I", ObjectId._inc)[1:4]
        ObjectId._inc = (ObjectId._inc + 1) % (0xFFFFFF + 1)
    return ObjectId(oid)


async def __local_restore():
    # restore user
    dot_rethink_path = config.get_settings().LOCAL_STORAGE_PATH / ".data" / ".rethink.json"
    if not dot_rethink_path.exists():
        return
    with open(dot_rethink_path, "r", encoding="utf-8") as f:
        u_insertion = json.load(f)

    u: UserMeta = {
        "_id": ObjectId(u_insertion["_id"]),
        "id": u_insertion["id"],
        "email": u_insertion["email"],
        "nickname": u_insertion["nickname"],
        "avatar": u_insertion["avatar"],
        "account": u_insertion["email"],
        "language": u_insertion["language"],

        "source": const.UserSource.LOCAL.value,
        "hashed": "",
        "disabled": False,
        "modifiedAt": datetime.datetime.now(tz=utc),
        "usedSpace": 0,
        "type": const.USER_TYPE.NORMAL.id,
        "lastState": {
            "recentCursorSearchSelectedNIds": [],
            "recentSearch": [],
            "nodeDisplayMethod": const.NodeDisplayMethod.CARD.value,
            "nodeDisplaySortKey": "modifiedAt"
        }
    }
    _ = await COLL.users.insert_one(u)

    # restore nodes
    md_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "md"
    if not md_dir.exists():
        return
    ns = []
    search_docs = []
    for md_path in md_dir.glob("*.md"):
        md = md_path.read_text(encoding="utf-8")
        created_time = datetime.datetime.fromtimestamp(md_path.stat().st_ctime, tz=utc)
        modified_time = datetime.datetime.fromtimestamp(md_path.stat().st_mtime, tz=utc)
        title_, body_, snippet_ = utils.preprocess_md(md)

        n: Node = {
            "_id": __oid_from_datetime(created_time),
            "id": md_path.stem,
            "uid": u_insertion["id"],
            "title": title_,
            "snippet": snippet_,
            "md": md,
            "type": const.NodeType.MARKDOWN.value,
            "disabled": False,
            "inTrash": False,
            "modifiedAt": modified_time,
            "inTrashAt": None,
            "fromNodeIds": [],
            "toNodeIds": [],
        }
        ns.append(n)
        search_docs.append(
            SearchDoc(
                nid=n["id"],
                title=title_,
                body=body_,
            )
        )
    res = await COLL.nodes.insert_many(ns)
    if not res.acknowledged:
        raise ValueError("cannot insert default node")
    logger.info(f"restore nodes count: {len(res.inserted_ids)}")

    docs = []
    for f in config.get_settings().LOCAL_STORAGE_PATH.glob(".data/files/*"):
        filename = f.name
        fid = filename.split(".")[0]
        created_time = datetime.datetime.fromtimestamp(f.stat().st_ctime, tz=utc)
        docs.append({
            "_id": __oid_from_datetime(created_time),
            "uid": u_insertion["id"],
            "fid": fid,
            "filename": filename,
            "size": f.stat().st_size,
        })
    res = await COLL.user_file.insert_many(docs)
    if not res.acknowledged:
        raise ValueError("cannot insert default node")

    await SEARCHER.drop()
    await SEARCHER.init()
    await SEARCHER.add_batch(
        uid=u["id"],
        docs=search_docs,
    )
    logger.info(f"restore files count: {len(res.inserted_ids)}")


async def __local_try_create_or_restore():
    if not config.get_settings().ONE_USER:
        return

    # check if field changes
    for c, t in [
        (COLL.users, UserMeta),
        (COLL.nodes, Node),
        (COLL.user_file, UserFile),
        (COLL.import_data, ImportData),
    ]:
        doc = await c.find_one()
        if doc:
            if set(doc.keys()) != set(t.__annotations__):
                await drop_all()
                await __local_restore()
                return
    # check local files count matches db
    md_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "md"
    if md_dir.exists():
        if len(list(md_dir.glob("*.md"))) != await COLL.nodes.count_documents({}):
            await drop_all()
            await __local_restore()
            return
    files_dir = config.get_settings().LOCAL_STORAGE_PATH / ".data" / "files"
    if files_dir.exists():
        if len(list(files_dir.glob("*"))) != await COLL.user_file.count_documents({}):
            await drop_all()
            await __local_restore()
            return

    # try fix TypeError: can't compare offset-naive and offset-aware datetimes
    docs = COLL.nodes.find()
    for doc in await docs.to_list(length=None):
        await COLL.nodes.update_one(
            {"_id": doc["_id"]},
            {"$set": {"modifiedAt": doc["modifiedAt"].replace(tzinfo=utc)}},
        )
    docs = COLL.import_data.find()
    for doc in await docs.to_list(length=None):
        await COLL.import_data.update_one(
            {"_id": doc["_id"]},
            {"$set": {"startAt": doc["startAt"].replace(tzinfo=utc)}},
        )

    if await COLL.users.find_one(
            {"account": const.DEFAULT_USER["email"], "source": const.UserSource.LOCAL.value}
    ) is not None:
        return

    # no default user, create one
    await __local_try_add_default_user()


async def __try_restore_search():
    count_mongo = await COLL.nodes.count_documents({})
    count_search = await SEARCHER.count_all()
    if count_mongo == count_search:
        return
    await SEARCHER.drop()
    await SEARCHER.init()
    docs = await COLL.nodes.find({}).to_list(length=None)
    search_docs = {}
    for doc in docs:
        if doc["uid"] not in search_docs:
            search_docs[doc["uid"]] = []

        search_docs[doc["uid"]].append(
            RestoreSearchDoc(
                nid=doc["id"],
                title=doc["title"],
                body=doc["md"],
                createdAt=doc["_id"].generation_time,
                modifiedAt=doc["modifiedAt"],
                disabled=doc["disabled"],
                inTrash=doc["inTrash"],
            )
        )
    for uid, docs in search_docs.items():
        code = await SEARCHER.batch_restore_docs(
            uid=uid,
            docs=docs,
        )
        if code != const.Code.OK:
            raise ValueError("cannot restore search index")
    logger.info(f"restore search index count: {count_mongo}")



from typing import List, Sequence, Literal

from pydantic import Field, BaseModel, NonNegativeInt

from rethink import const


class NodesSearchResponse(BaseModel):
    class Data(BaseModel):
        class Node(BaseModel):
            id: str
            title: str
            snippet: str
            titleHighlight: str
            bodyHighlights: List[str]
            score: float
            type: int
            createdAt: str
            modifiedAt: str

        nodes: List[Node]
        total: NonNegativeInt

    code: NonNegativeInt
    message: str
    requestId: str
    data: Data


class SearchUserNodesRequest(BaseModel):
    query: str = Field(default="", max_length=const.SEARCH_QUERY_MAX_LENGTH)
    sortKey: Literal[
        "createdAt", "modifiedAt", "title", "similarity"
    ] = Field(default="createdAt", max_length=20)
    reverse: bool = False
    page: NonNegativeInt = 0
    pageSize: NonNegativeInt = 0
    nidExclude: Sequence[str] = Field(default_factory=list, max_items=1000)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class CursorQueryRequest(BaseModel):
    nid: str = Field(max_length=const.NID_MAX_LENGTH)
    query: str = Field(max_length=const.SEARCH_QUERY_MAX_LENGTH)
    page: NonNegativeInt = 0
    pageSize: NonNegativeInt = 0
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class RecommendNodesRequest(BaseModel):
    content: str = Field(max_length=const.RECOMMEND_CONTENT_MAX_LENGTH)
    nidExclude: Sequence[str] = Field(default_factory=list, max_items=1000)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class RecommendNodesResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    nodes: List[NodesSearchResponse.Data.Node]


class AddToRecentSearchHistRequest(BaseModel):
    nid: str = Field(max_length=const.NID_MAX_LENGTH)
    toNid: str = Field(max_length=const.NID_MAX_LENGTH)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class GetRecentSearchResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: str
    queries: List[str]



from typing import List, Dict, Tuple, Sequence, Literal

from rethink import const
from rethink.controllers.schemas.search import NodesSearchResponse
from rethink.controllers.utils import datetime2str
from rethink.models import tps
from rethink.models.database import COLL, searcher
from rethink.models.search_engine.engine import SearchResult


async def _2node_data(
        hits: Sequence[SearchResult],
) -> List[NodesSearchResponse.Data.Node]:
    nodes = await COLL.nodes.find({"id": {"$in": [hit.nid for hit in hits]}}).to_list(length=None)
    nodes_map: Dict[str, tps.Node] = {n["id"]: n for n in nodes}
    results = []
    for hit in hits:
        n = nodes_map[hit.nid]
        r = NodesSearchResponse.Data.Node(
            id=n["id"],
            title=n["title"],
            snippet=n["snippet"],
            titleHighlight=hit.titleHighlight,
            bodyHighlights=hit.bodyHighlights,
            score=hit.score,
            type=n["type"],
            createdAt=datetime2str(n["_id"].generation_time),
            modifiedAt=datetime2str(n["modifiedAt"]),
        )
        results.append(r)
    return results


async def search(
        uid: str,
        query: str,
        sort_key: Literal[
            "createdAt", "modifiedAt", "title", "similarity"
        ],
        reverse: bool,
        page: int,
        page_size: int,
        exclude_nids: Sequence[str],
) -> Tuple[List[NodesSearchResponse.Data.Node], int]:
    # search nodes
    hits, total = await searcher().search(
        uid=uid,
        query=query,
        sort_key=sort_key,
        reverse=reverse,
        page=page,
        page_size=page_size,
        exclude_nids=exclude_nids,
    )
    results = await _2node_data(hits)

    if query != "":
        await put_recent_search(uid, query)
    return results, total


async def recommend(
        uid: str,
        content: str,
        max_return: int = 5,
        exclude_nids: Sequence[str] = None,
) -> List[NodesSearchResponse.Data.Node]:
    if content == "":
        return []
    # search nodes
    hits = await searcher().recommend(
        uid=uid,
        content=content,
        max_return=max_return,
        exclude_nids=exclude_nids,
    )
    return await _2node_data(hits)


async def cursor_query(
        uid: str,
        nid: str,
        query: str,
        page: int,
        page_size: int,
) -> Tuple[List[NodesSearchResponse.Data.Node], int]:
    # if cursor_text.startswith("@"):
    #     query = cursor_text[1:].strip()
    # else:
    #     found = CURSOR_AT_PTN.search(cursor_text)
    #     if found is None:
    #         return None, []
    #     query = found.group(1).strip()

    query = query.strip()

    # if query == "", return recent nodes
    if query == "":
        u = await COLL.users.find_one({"id": uid})
        if u is None:
            return [], 0
        rn = u["lastState"]["recentCursorSearchSelectedNIds"]
        try:
            rn.remove(nid)
        except ValueError:
            pass
        nodes = sorted(
            await COLL.nodes.find({"id": {"$in": rn}}).to_list(length=None), key=lambda x: rn.index(x["id"])
        )
        return [
            NodesSearchResponse.Data.Node(
                id=n["id"],
                title=n["title"],
                snippet=n["snippet"][:60] + "...",
                titleHighlight=n["title"],
                bodyHighlights=[n["snippet"][:60] + "..."],
                score=0,
                type=n["type"],
                createdAt=datetime2str(n["_id"].generation_time),
                modifiedAt=datetime2str(n["modifiedAt"]),
            ) for n in nodes
        ], len(rn)

    return await search(
        uid=uid,
        query=query,
        sort_key="similarity",
        reverse=True,
        page=page,
        page_size=page_size,
        exclude_nids=[nid],
    )


async def add_recent_cursor_search(
        uid: str,
        nid: str,
        to_nid: str,
) -> const.Code:
    # add selected node to recentCursorSearchSelectedNIds
    user_c = {"id": uid, "disabled": False}
    node_c = {"uid": uid, "id": {"$in": [nid, to_nid]}}

    # try finding user
    u = await COLL.users.find_one(user_c)
    if u is None:
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR

    # try finding node
    ns = await COLL.nodes.find(node_c).to_list(length=None)
    if len(ns) != 2:
        return const.Code.NODE_NOT_EXIST

    rns = u["lastState"]["recentCursorSearchSelectedNIds"]
    if to_nid in rns:
        rns.remove(to_nid)
    rns.insert(0, to_nid)
    if len(rns) > 10:
        rns = rns[:10]

    # add to recentCursorSearchSelectedNIds
    res = await COLL.users.update_one(
        {"id": uid},
        {"$set": {"lastState.recentCursorSearchSelectedNIds": rns}}
    )
    if res.matched_count != 1:
        return const.Code.OPERATION_FAILED

    return const.Code.OK


async def get_recent_search(uid: str) -> List[str]:
    doc = await COLL.users.find_one({"id": uid})
    if doc is None:
        return []
    return doc["lastState"]["recentSearch"]


async def put_recent_search(uid: str, query: str) -> const.Code:
    doc = await COLL.users.find_one({"id": uid})
    if doc is None:
        return const.Code.ACCOUNT_OR_PASSWORD_ERROR
    rns = doc["lastState"]["recentSearch"]
    try:
        rns.remove(query)
    except ValueError:
        pass
    rns.insert(0, query)
    if len(rns) > 20:
        rns = rns[:20]
    _ = await COLL.users.update_one(
        {"id": uid},
        {"$set": {"lastState.recentSearch": rns}}
    )
    return const.Code.OK



import datetime
import io
import ipaddress
import math
import re
import socket
import uuid
from html.parser import HTMLParser
from io import StringIO
from typing import Tuple
from urllib.parse import urlparse

import httpx
import jwt
from markdown import Markdown

from rethink import config, const
from rethink.logger import logger

HEADERS = {
    'typ': 'jwt',
    'alg': 'RS256'
}
alphabet = "3467ACDEFGHJKLMNPQRTUVWXYabcdefghkmnoprtuvwxy"
alphabet_len = len(alphabet)
__padding = int(math.ceil(math.log(2 ** 128, alphabet_len)))
__code_pattern = re.compile(r"^```[^\S\r\n]*[a-z]*?\n(.*?)\n```$", re.MULTILINE | re.DOTALL)


def short_uuid() -> str:
    """
    The output has the most significant digit first.
    """
    number = uuid.uuid4().int
    output = ""

    while number:
        number, digit = divmod(number, alphabet_len)
        output += alphabet[digit]
    if __padding:
        remainder = max(__padding - len(output), 0)
        output = output + alphabet[0] * remainder
    return output[::-1]


def __unmark_element(element, stream=None):
    if stream is None:
        stream = io.StringIO()
    if element.text:
        stream.write(element.text)
    for sub in element:
        __unmark_element(sub, stream)
    if element.tail:
        stream.write(element.tail)
    return stream.getvalue()


# patching Markdown
Markdown.output_formats["plain"] = __unmark_element
__md = Markdown(output_format="plain")
__md.stripTopLevelTags = False


def md2txt(md: str) -> str:
    for found in list(__code_pattern.finditer(md))[::-1]:
        span = found.span()
        code = found.group(1)
        md = f"{md[: span[0]]}{code}{md[span[1]:]}"
    return __md.convert(md)


def preprocess_md(md: str, snippet_len: int = 200) -> Tuple[str, str, str]:
    title, body = split_title_body(fulltext=md)
    title = md2txt(title.strip())
    body = md2txt(body.strip())
    snippet = strip_html_tags(body)[:snippet_len]
    return title, body, snippet


def split_title_body(fulltext: str) -> (str, str):
    title_body = fulltext.split("\n", maxsplit=1)
    title = title_body[0].strip()
    try:
        body = title_body[1].strip()
    except IndexError:
        body = ""
    return title, body


def jwt_encode(exp_delta: datetime.timedelta, data: dict) -> str:
    """
    Create token

    Args:
        exp_delta: expired delta
        data: payload data

    Returns:
        str: token
    """
    payload = {
        "exp": datetime.datetime.utcnow() + exp_delta
    }
    payload.update(data)
    token = jwt.encode(
        payload=payload,
        key=config.get_settings().JWT_KEY,
        algorithm=HEADERS["alg"],
        headers=HEADERS,
    )
    return token


def jwt_decode(token: str) -> dict:
    return jwt.decode(
        token,
        key=config.get_settings().JWT_KEY_PUB,
        algorithms=[HEADERS["alg"]],
        options={"verify_exp": True}
    )


def change_link_title(md: str, nid: str, new_title: str) -> str:
    new_md = re.sub(
        r"\[@[^].]*?]\(/n/{}/?\)".format(nid),
        f"[@{new_title}](/n/{nid})",
        md,
    )
    return new_md


ONLY_HTTP_LINK_PTN = re.compile(r"^https?://\S*$")


def contain_only_http_link(md: str) -> str:
    content = md.strip()
    if ONLY_HTTP_LINK_PTN.match(content) is None:
        return ""
    return content


ASYNC_CLIENT_HEADERS = {
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,"
              "application/signed-exchange;v=b3;q=0.7",
    "Accept-Encoding": "gzip, deflate, br",
    "Accept-Language": "zh",
    "Cache-Control": "no-cache",
    "Pragma": "no-cache",
    "Sec-Ch-Ua": '"Chromium";v="118", "Google Chrome";v="118", "Not=A?Brand";v="99"',
    "Sec-Ch-Ua-Mobile": "?0",
    "Sec-Ch-Ua-Platform": '"macOS"',
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "none",
    "Sec-Fetch-User": "?1",
    "Upgrade-Insecure-Requests": "1",
    "User-Agent": "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) "
                  "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/118.0.0.0 Safari/537.36",
}


def is_internal_ip(ip):
    internal_ip_ranges = [
        ("127.0.0.0", "127.255.255.255"),
        ("10.0.0.0", "10.255.255.255"),
        ("9.0.0.0", "9.255.255.255"),
        ("100.64.0.0", "100.127.255.255"),
        ("192.168.0.0", "192.168.255.255"),
        ("172.16.0.0", "172.31.255.255"),
    ]

    for start, end in internal_ip_ranges:
        if ipaddress.IPv4Address(start) <= ip <= ipaddress.IPv4Address(end):
            return True
    return False


# SSRF protection
def ssrf_check(url: str) -> bool:
    if url.startswith("http://localhost") or url.startswith("http://127.0.0.1"):
        return True
    if '@' in url:
        return True
    host = urlparse(url).hostname
    try:
        ip = ipaddress.IPv4Address(socket.gethostbyname(host))
    except socket.gaierror:
        return True
    if ip.is_private:
        return True
    if is_internal_ip(ip):
        return True


async def get_title_description_from_link(url: str, language: str) -> Tuple[str, str]:
    if language == const.Language.ZH.value:
        no_title = "网址没发现标题"
        no_description = "网址没发现描述"
    elif language == const.Language.EN.value:
        no_title = "No title found"
        no_description = "No description found"
    else:
        no_title = "No title found"
        no_description = "No description found"

    # SSRF protection
    if ssrf_check(url):
        return no_title, no_description
    # end of SSRF protection

    async with httpx.AsyncClient() as client:
        try:
            response = await client.get(
                url=url,
                headers=ASYNC_CLIENT_HEADERS,
                timeout=3.
            )
        except (
                httpx.ConnectTimeout,
                RuntimeError,
                httpx.ConnectError,
                httpx.ReadTimeout,
                httpx.HTTPError
        ) as e:
            logger.info(f"failed to get {url}: {e}")
            return no_title, no_description
        if response.status_code in (301, 302):
            url = response.headers["Location"]
            if ssrf_check(url):
                return no_title, no_description
            return await get_title_description_from_link(url=url, language=language)
        if response.status_code != 200:
            return no_title, no_description
        html = response.text

    title, description = "", ""
    found = re.search(r'<meta[^>]*name="title"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="title"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r"<title[^>]*>(.*?)</title>", html, re.DOTALL)
    if found:
        title = found.group(1).strip()

    found = re.search(r'<meta[^>]*name="description"[^>]*content="([^"]*)"[^>]*>', html, re.DOTALL)
    if found is None:
        found = re.search(r'<meta[^>]*content="([^"]*)"[^>]*name="description"[^>]*>', html, re.DOTALL)
    if found:
        description = found.group(1).strip()[:400]
    if title == "":
        title = no_title
    if description == "":
        description = no_description
    return title, description


class MLStripper(HTMLParser):
    def __init__(self):
        super().__init__()
        self.reset()
        self.strict = False
        self.convert_charrefs = True
        self.text = StringIO()

    def handle_data(self, d):
        self.text.write(d)

    def get_data(self):
        return self.text.getvalue()


def strip_html_tags(html):
    s = MLStripper()
    s.feed(html)
    return s.get_data()


def mask_email(email: str):
    if email == "":
        return ""
    if "@" not in email:
        return email
    name, end = email.split("@", 1)
    if len(name) <= 1:
        e = f"{name[0]}**@{end}"
    elif len(name) == 2:
        e = f"{name[0]}**{name[1]}@{end}"
    else:
        e = f"{name[:2]}**{name[-1]}@{end}"
    return e



from rethink import const, models
from rethink.controllers import schemas
from rethink.controllers.utils import TokenDecode


async def cursor_query(
        td: TokenDecode,
        req: schemas.search.CursorQueryRequest,
) -> schemas.search.NodesSearchResponse:
    if td.code != const.Code.OK:
        return schemas.search.NodesSearchResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            data=None
        )
    nodes, total = await models.search.cursor_query(
        uid=td.uid,
        nid=req.nid,
        query=req.query,
        page=req.page,
        page_size=req.pageSize,
    )
    code = const.Code.OK
    return schemas.search.NodesSearchResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        data=schemas.search.NodesSearchResponse.Data(
            nodes=nodes,
            total=total,
        )
    )


async def search_user_nodes(
        td: TokenDecode,
        req: schemas.search.SearchUserNodesRequest,
) -> schemas.search.NodesSearchResponse:
    if td.code != const.Code.OK:
        return schemas.search.NodesSearchResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            nodes=[],
        )
    nodes, total = await models.search.search(
        uid=td.uid,
        query=req.query,
        sort_key=req.sortKey,
        reverse=req.reverse,
        page=req.page,
        page_size=req.pageSize,
        exclude_nids=req.nidExclude,
    )
    code = const.Code.OK
    return schemas.search.NodesSearchResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        data=schemas.search.NodesSearchResponse.Data(
            nodes=nodes,
            total=total,
        )
    )


async def recommend_nodes(
        td: TokenDecode,
        req: schemas.search.RecommendNodesRequest,
) -> schemas.search.RecommendNodesResponse:
    if td.code != const.Code.OK:
        return schemas.search.RecommendNodesResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
            nodes=[],
        )
    max_return = 5
    nodes = await models.search.recommend(
        uid=td.uid,
        content=req.content,
        max_return=max_return,
        exclude_nids=req.nidExclude,
    )
    code = const.Code.OK
    return schemas.search.RecommendNodesResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
        nodes=nodes
    )


async def add_to_recent_cursor_search(
        td: TokenDecode,
        req: schemas.search.AddToRecentSearchHistRequest
) -> schemas.base.AcknowledgeResponse:
    if td.code != const.Code.OK:
        return schemas.base.AcknowledgeResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=req.requestId,
        )
    code = await models.search.add_recent_cursor_search(uid=td.uid, nid=req.nid, to_nid=req.toNid)
    return schemas.base.AcknowledgeResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=req.requestId,
    )


async def get_recent(
        td: TokenDecode,
        rid: str,
) -> schemas.search.GetRecentSearchResponse:
    if td.code != const.Code.OK:
        return schemas.search.GetRecentSearchResponse(
            code=td.code.value,
            message=const.get_msg_by_code(td.code, td.language),
            requestId=rid,
            queries=[],
        )
    queries = await models.search.get_recent_search(uid=td.uid)
    code = const.Code.OK
    return schemas.search.GetRecentSearchResponse(
        code=code.value,
        message=const.get_msg_by_code(code, td.language),
        requestId=rid,
        queries=queries,
    )



import base64
import hashlib
import traceback
from typing import Optional, Tuple

import bcrypt
import jwt
from fastapi import Header

from rethink import config, const, models, regex
from rethink.controllers.utils import TokenDecode
from rethink.logger import logger
from rethink.models.utils import jwt_decode


def __one_line_traceback() -> str:
    return traceback.format_exc().replace("\n", "\\n")


async def token2uid(token: str = Header(...)) -> TokenDecode:
    uid = ""
    err = ""
    language = const.Language.EN.value
    try:
        payload = jwt_decode(token=token)
        u, code = await models.user.get(uid=payload["uid"])
        if code != const.Code.OK:
            logger.error(f"models.user.get err: {const.CODE_MESSAGES[code].zh}")
            return TokenDecode(code=code, language=language)
        uid = u["id"]
        language = u["language"]
    except jwt.exceptions.ExpiredSignatureError:
        code = const.Code.EXPIRED_AUTH
        err = "auth token expired"
    except jwt.exceptions.DecodeError:
        code = const.Code.INVALID_AUTH
        err = __one_line_traceback()
    except jwt.exceptions.InvalidTokenError:
        code = const.Code.INVALID_AUTH
        err = __one_line_traceback()
    except Exception:  # pylint: disable=broad-except
        code = const.Code.INVALID_AUTH
        err = __one_line_traceback()
    if code != const.Code.OK:
        logger.error(f"jwt_decode err: {err}")
    return TokenDecode(code=code, uid=uid, language=language)


async def get_user_by_email(email: str) -> Tuple[Optional[models.tps.UserMeta], const.Code]:
    return await models.user.get_by_email(email=email)


async def verify_user(u: models.tps.UserMeta, password: str) -> bool:
    if config.get_settings().ONE_USER:
        return True
    base_pw = _base_password(password=password, email=u["email"])
    db_hash = await models.user.get_hash_by_uid(u["id"])
    match = bcrypt.checkpw(base_pw, db_hash.encode("utf-8"))
    return match


def _base_password(password: str, email: str) -> bytes:
    return base64.b64encode(hashlib.sha256(f"{password}&&{email}".encode("utf-8")).digest())


def validate_email_pwd(email: str, password: str) -> const.Code:
    if config.get_settings().ONE_USER:
        logger.warning("on ONE_USER mode, user registration will be skipped")
        return const.Code.ONE_USER_MODE
    if regex.EMAIL_PTN.match(email) is None:
        return const.Code.INVALID_EMAIL
    if regex.VALID_PASSWORD_PTN.match(password) is None:
        return const.Code.INVALID_PASSWORD
    return const.Code.OK


async def register_user(
        email: str,
        password: str,
        language: str = const.Language.EN.value,
) -> Tuple[str, const.Code]:
    code = validate_email_pwd(email=email, password=password)
    if code != const.Code.OK:
        return "", code
    u, code = await models.user.get_by_email(email=email)
    if code == const.Code.OK or u is not None:
        return "", const.Code.USER_EXIST

    bpw = _base_password(password=password, email=email)
    hashed = bcrypt.hashpw(bpw, config.get_settings().DB_SALT)
    uid, code = await models.user.add(
        account=email,
        source=const.UserSource.EMAIL.value,
        email=email,
        hashed=hashed.decode("utf-8"),
        nickname=email.split("@")[0],
        avatar="",
        language=language,
    )
    if code != const.Code.OK:
        return "", code

    code = await models.node.new_user_add_default_nodes(uid=uid, language=language)
    return uid, code


async def reset_password(
        email: str,
        password: str,
) -> Tuple[Optional[models.tps.UserMeta], const.Code]:
    code = validate_email_pwd(email=email, password=password)
    if code != const.Code.OK:
        return None, code
    u, code = await models.user.get_by_email(email=email)
    if code != const.Code.OK:
        return None, code
    if u is None:
        return None, const.Code.INVALID_AUTH
    bpw = _base_password(password=password, email=email)
    hashed = bcrypt.hashpw(bpw, config.get_settings().DB_SALT)
    code = await models.user.reset_password(uid=u["id"], hashed=hashed.decode("utf-8"))
    return u, code



import unittest

from rethink import models, const
from . import utils


class DataRestoreTest(unittest.IsolatedAsyncioTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.local")

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.local")

    async def asyncSetUp(self) -> None:
        await models.database.drop_all()

    async def asyncTearDown(self) -> None:
        await models.database.drop_all()

    async def test_restore_search(self):
        await models.database.init()
        u, _ = await models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.uid = u["id"]
        base_count = 2
        nids = []
        for i in range(20):
            n, code = await models.node.add(
                uid=self.uid,
                md=f"title{i}\ntext{i}",
                type_=const.NodeType.MARKDOWN.value,
            )
            nids.append(n["id"])
            self.assertEqual(const.Code.OK, code)
        self.assertEqual(20 + base_count, await models.database.searcher().count_all())

        code = await models.node.batch_to_trash(uid=self.uid, nids=nids[:10])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(20 + base_count, await models.database.searcher().count_all())

        code = await models.database.searcher().delete_batch(
            uid=self.uid,
            nids=nids[:10],
        )
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(10 + base_count, await models.database.searcher().count_all())

        await models.database.init()
        self.assertEqual(20 + base_count, await models.database.searcher().count_all())



import datetime
from dataclasses import dataclass, field
from typing import Sequence
from urllib.parse import urlparse

from rethink import const


@dataclass
class TokenDecode:
    code: const.Code
    uid: str = ""
    language: str = field(default=const.Language.EN.value)


def datetime2str(dt: datetime.datetime) -> str:
    return dt.strftime('%Y-%m-%dT%H:%M:%SZ')


def is_allowed_mime_type(data_url, allowed_mime_types: Sequence[str]):
    try:
        result = urlparse(data_url)
        if result.scheme != 'data':
            return False

        media_type = result.path.split(';')[0]
        return media_type in allowed_mime_types
    except Exception:  # pylint: disable=broad-except
        return False



import multiprocessing
import shutil
import socket
import time
import unittest
import urllib.request
from pathlib import Path

from rethink import run, config


class TestRun(unittest.TestCase):
    path: Path

    @classmethod
    def setUpClass(cls) -> None:
        config.get_settings.cache_clear()
        cls.path = p = Path(__file__).parent / "tmp"
        p.mkdir(exist_ok=True)

    @classmethod
    def tearDownClass(cls) -> None:
        shutil.rmtree(str(cls.path), ignore_errors=True)
        config.get_settings.cache_clear()

    def test_run(self):
        port = 8001

        for _ in range(2):
            p = multiprocessing.Process(target=run, kwargs={"path": self.path, "port": port, "language": "zh"})
            p.start()
            # p.join()
            while True:
                sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
                sock.settimeout(2)
                result = sock.connect_ex(('127.0.0.1', port))
                sock.close()
                if result == 0:
                    break
                time.sleep(0.1)

            for url in [
                "",
                "/login",
                "/about",
                "/r",
                "/r/settings",
                "/r/user",
                "/r/import",
                "/n/123",
            ]:
                resp = urllib.request.urlopen(f"http://127.0.0.1:{port}{url}")
                self.assertEqual(200, resp.status, msg=f"failed to get {url}")
                self.assertEqual("text/html; charset=utf-8", resp.headers["content-type"])
                self.assertIn("Rethink", resp.read().decode())

            p.kill()
            p.join()
            self.assertTrue(self.path.exists())
            self.assertTrue(self.path.is_dir())
            self.assertTrue((self.path / ".data").exists())
            self.assertEqual(2, len(list((self.path / ".data" / "md").glob("*.md"))))



import unittest
from textwrap import dedent

import elastic_transport
import pymongo.errors
from bson import ObjectId

from rethink import const, models, config
from rethink.controllers.auth import register_user
from . import utils


class RemoteModelsTest(unittest.IsolatedAsyncioTestCase):
    default_pwd = "rethink123"

    @classmethod
    def setUpClass(cls) -> None:
        utils.set_env(".env.test.development")

    @classmethod
    def tearDownClass(cls) -> None:
        utils.drop_env(".env.test.development")

    async def asyncSetUp(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncSetUp skipped")
            return
        try:
            await models.database.drop_all()
            await models.database.init()
            uid, code = await register_user(
                email=const.DEFAULT_USER["email"],
                password=self.default_pwd,
                language=const.Language.EN.value)
            self.assertEqual(const.Code.OK, code)
            self.token = models.utils.jwt_encode(
                exp_delta=config.get_settings().JWT_EXPIRED_DELTA,
                data={"uid": uid, "language": const.Language.EN.value},
            )
            self.uid = uid
        except (
                pymongo.errors.NetworkTimeout,
                pymongo.errors.ServerSelectionTimeoutError,
                elastic_transport.ConnectionError,
                RuntimeError,
        ):
            print("remote test asyncSetUp timeout")
            utils.skip_no_connect.skip = True

    async def asyncTearDown(self) -> None:
        if utils.skip_no_connect.skip:
            print("remote test asyncTearDown skipped")
            return
        try:
            await models.database.drop_all()
        except (
                pymongo.errors.NetworkTimeout,
                pymongo.errors.ServerSelectionTimeoutError,
                elastic_transport.ConnectionError,
                RuntimeError,
        ):
            print("remote test asyncTearDown timeout")
            utils.skip_no_connect.skip = True

    @utils.skip_no_connect
    async def test_user(self):
        if models.database.CLIENT is None: return
        _id, code = await register_user(email="aaa@rethink.run", password="bbb123")
        self.assertNotEqual("", str(_id))
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get_by_email(email=const.DEFAULT_USER["email"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("rethink", u["nickname"])
        self.assertIsNotNone(u)

        u, code = await models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("aaa", u["nickname"])

        u, code = await models.user.update(uid=_id, hashed="1", nickname="2", avatar="3")
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get(_id)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("1", u["hashed"])
        self.assertEqual("2", u["nickname"])
        self.assertEqual("3", u["avatar"])

        code = await models.user.disable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        u, code = await models.user.get(uid=_id)
        self.assertEqual(const.Code.ACCOUNT_OR_PASSWORD_ERROR, code)

        code = await models.user.enable(uid=_id)
        self.assertEqual(const.Code.OK, code)

        code = await models.user.disable(uid="ssaa")
        self.assertEqual(const.Code.OPERATION_FAILED, code)

        code = await models.user.delete(uid=_id)
        self.assertEqual(const.Code.OK, code)

    @utils.skip_no_connect
    async def test_node(self):
        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        used_space = u["usedSpace"]
        node, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + len(node["md"].encode("utf-8")), u["usedSpace"])

        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title", n["title"])

        await models.database.searcher().refresh()
        ns, total = await models.database.searcher().search(uid=self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        used_space = u["usedSpace"]
        n, code = await models.node.update(uid=self.uid, nid=node["id"], md="# title2\ntext2")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual("title2", n["title"])
        self.assertEqual("text2", n["snippet"])
        self.assertEqual(const.NodeType.MARKDOWN.value, n["type"])

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space + (
                len(n["md"].encode("utf-8")) -
                len(node["md"].encode("utf-8"))
        ), u["usedSpace"])

        code = await models.node.disable(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        await models.database.searcher().refresh()
        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        code = await models.node.to_trash(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        await models.database.searcher().refresh()
        code = await models.node.delete(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertIsNone(n)
        self.assertEqual(const.Code.NODE_NOT_EXIST, code)

        u, code = await models.user.get(self.uid)
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(used_space - len(node["md"].encode("utf-8")), u["usedSpace"])

    @utils.skip_no_connect
    async def test_parse_at(self):
        nid1, _ = await models.node.add(
            uid=self.uid, md="c", type_=const.NodeType.MARKDOWN.value,
        )
        nid2, _ = await models.node.add(
            uid=self.uid, md="我133", type_=const.NodeType.MARKDOWN.value,
        )
        md = dedent(
            f"""title
            fffqw [@c](/n/{nid1['id']})
            fff
            [@我133](/n/{nid2['id']})
            ffq
            """)
        node, code = await models.node.add(
            uid=self.uid, md=md, type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        await models.database.searcher().refresh()
        nodes, total = await models.database.searcher().search(uid=self.uid)
        self.assertEqual(5, len(nodes))
        self.assertEqual(5, total)

        found, total = await models.database.searcher().search(uid=self.uid, query="我")
        self.assertEqual(2, len(found), msg=found)
        self.assertEqual(2, total)

        recommend = await models.database.searcher().recommend(
            uid=self.uid,
            content="I do need a Knowledge Management System. This is a good one to try.",
            exclude_nids=[],
        )
        self.assertEqual(1, len(recommend))

        n, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(2, len(n["toNodeIds"]))

        tmp_text = n["md"]
        n, code = await models.node.update(uid=self.uid, nid=node["id"], md=tmp_text + "xxxx")
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(tmp_text + "xxxx", n["md"])

        n, code = await models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(n["fromNodeIds"]))

        n, code = await models.node.update(uid=self.uid, nid=node["id"], md=n["md"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["toNodeIds"]))

        n, code = await models.node.get(uid=self.uid, nid=nid1['id'])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(0, len(n["fromNodeIds"]))

    @utils.skip_no_connect
    async def test_add_set(self):
        node, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(0, len(node["toNodeIds"]))
        self.assertEqual(const.Code.OK, code)

        res = await models.db_ops.node_add_to_set(node["id"], "toNodeIds", ObjectId())
        self.assertEqual(1, res.modified_count)
        node, code = await models.node.get(uid=self.uid, nid=node["id"])
        self.assertEqual(const.Code.OK, code)
        self.assertEqual(1, len(node["toNodeIds"]))

    @utils.skip_no_connect
    async def test_to_trash(self):
        n1, code = await models.node.add(
            uid=self.uid, md="title\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)
        n2, code = await models.node.add(
            uid=self.uid, md="title2\ntext", type_=const.NodeType.MARKDOWN.value
        )
        self.assertEqual(const.Code.OK, code)

        code = await models.node.to_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        ns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(1, len(ns))
        self.assertEqual(1, total)
        self.assertEqual(n1["id"], ns[0]["id"])

        await models.database.searcher().refresh()
        ns, total = await models.database.searcher().search(self.uid)
        self.assertEqual(3, len(ns))
        self.assertEqual(3, total)

        code = await models.node.restore_from_trash(self.uid, n1["id"])
        self.assertEqual(const.Code.OK, code)

        await models.database.searcher().refresh()
        nodes, total = await models.database.searcher().search(self.uid)
        self.assertEqual(4, len(nodes))
        self.assertEqual(4, total)

    @utils.skip_no_connect
    async def test_batch(self):
        ns = []
        for i in range(10):
            n, code = await models.node.add(
                uid=self.uid, md=f"title{i}\ntext", type_=const.NodeType.MARKDOWN.value
            )
            self.assertEqual(const.Code.OK, code)
            ns.append(n)

        base_count = 2

        code = await models.node.batch_to_trash(self.uid, [n["id"] for n in ns[:4]])
        self.assertEqual(const.Code.OK, code)

        await models.database.searcher().refresh()
        nodes, total = await models.database.searcher().search(self.uid)
        self.assertEqual(6 + base_count, len(nodes))
        self.assertEqual(6 + base_count, total)

        tns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(4, total)
        self.assertEqual(4, len(tns))

        code = await models.node.restore_batch_from_trash(self.uid, [n["id"] for n in tns[:2]])
        self.assertEqual(const.Code.OK, code)

        await models.database.searcher().refresh()
        nodes, total = await models.database.searcher().search(self.uid)
        self.assertEqual(8 + base_count, len(nodes))
        self.assertEqual(8 + base_count, total)

        code = await models.node.batch_delete(self.uid, [n["id"] for n in tns[2:4]])
        self.assertEqual(const.Code.OK, code)

        tns, total = await models.node.get_nodes_in_trash(self.uid, 0, 10)
        self.assertEqual(0, total)
        self.assertEqual(0, len(tns))

    @utils.skip_no_connect
    async def test_update_used_space(self):
        u, code = await models.user.get(self.uid)
        base_used_space = u["usedSpace"]
        for delta, value in [
            (100, 100),
            (100, 200),
            (0, 200),
            (-3000, 0),
            (20.1, 20.1),
        ]:
            code = await models.user.update_used_space(self.uid, delta)
            self.assertEqual(const.Code.OK, code, msg=f"delta: {delta}, value: {value}")
            u, code = await models.user.get(self.uid)
            self.assertEqual(const.Code.OK, code)
            now = u["usedSpace"] - base_used_space
            if now < 0:
                now = 0
                base_used_space = 0
            self.assertAlmostEqual(value, now, msg=f"delta: {delta}, value: {value}")



from typing import Optional

from pydantic import BaseModel, NonNegativeInt


class AcknowledgeResponse(BaseModel):
    code: int
    message: str
    requestId: str


class TokenResponse(BaseModel):
    code: NonNegativeInt
    message: str
    requestId: Optional[str]
    token: str = ""



from typing import Literal

from pydantic import BaseModel, NonNegativeInt, HttpUrl, EmailStr, AfterValidator, Field
from typing_extensions import Annotated

from rethink import const


def empty_str_to_http_url(v: str) -> HttpUrl:
    if v == "":
        return v
    return HttpUrl(v)


class UserInfoResponse(BaseModel):
    class User(BaseModel):
        class LastState(BaseModel):
            nodeDisplayMethod: NonNegativeInt
            nodeDisplaySortKey: str

        email: str
        nickname: str
        avatar: Annotated[str, AfterValidator(empty_str_to_http_url)]
        createdAt: str
        language: Literal["en", "zh"]
        usedSpace: NonNegativeInt = 0
        maxSpace: NonNegativeInt = 0
        lastState: LastState

    code: NonNegativeInt
    message: str
    requestId: str
    user: User = None


class RegisterRequest(BaseModel):
    email: EmailStr = Field(max_length=const.EMAIL_MAX_LENGTH)
    password: str = Field(max_length=const.PASSWORD_MAX_LENGTH)
    captchaToken: str = Field(max_length=2000)
    captchaCode: str = Field(max_length=10)
    language: Literal["en", "zh"]
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class LoginRequest(BaseModel):
    email: EmailStr = Field(max_length=const.EMAIL_MAX_LENGTH)
    password: str = Field(max_length=const.PASSWORD_MAX_LENGTH)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class UpdateRequest(BaseModel):
    nickname: str = Field(default="", max_length=const.NICKNAME_MAX_LENGTH)
    avatar: str = Field(default="", max_length=2048)
    language: Literal["en", "zh"] = ""
    nodeDisplayMethod: int = Field(default=-1, ge=-1, le=10)
    nodeDisplaySortKey: Literal["modifiedAt", "createdAt", "title", ""] = Field(default="", max_length=20)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class EmailVerificationRequest(BaseModel):
    email: str = Field(max_length=const.EMAIL_MAX_LENGTH)
    captchaToken: str = Field(max_length=2000)
    captchaCode: str = Field(max_length=10)
    language: Literal["en", "zh"] = const.Language.EN.value
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)


class ResetPasswordRequest(BaseModel):
    email: str = Field(max_length=const.EMAIL_MAX_LENGTH)
    newPassword: str = Field(max_length=const.PASSWORD_MAX_LENGTH)
    verification: str = Field(max_length=const.PASSWORD_MAX_LENGTH)
    verificationToken: str = Field(max_length=2000)
    requestId: str = Field(default="", max_length=const.REQUEST_ID_MAX_LENGTH)



from fastapi.responses import StreamingResponse

from rethink.models.verify.verification import random_captcha


def get_captcha_img():
    token, data = random_captcha(length=4, sound=False)
    return StreamingResponse(
        data["img"],
        headers={
            "X-Captcha-Token": token
        },
        media_type="image/png",
    )
